/*****
  command line parser -- generated by clig (Version: |Version|)

  The command line parser `clig':
  (C) 1995 Harald Kirsch (kir@iitb.fhg.de)
*****/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* mainly a hack to make things work on SUNOS with gcc */
extern double strtod(const char *, char **);
extern long int strtol(const char *, char **, int);

#include <cmdline.h>

char *Program;

static float fDefault[] = {2, 4, 8, 16.11};
static int bDefault[] = {-1, 0, 1};

static Cmdline cmd = {
  /***** -v: switch on verbose program operation */
  /* verboseP = */ 0,
  /***** -title: title of x/y-plot */
  /* titleP = */ 0,
  /* title = */ (char*)0,
  /* titleC = */ 0,
  /***** -xytitles: title of x- and y-axis */
  /* xytitlesP = */ 0,
  /* xytitles = */ (char**)0,
  /* xytitlesC = */ 0,
  /***** -plotnames: names of curves to plot */
  /* plotnamesP = */ 0,
  /* plotnames = */ (char**)0,
  /* plotnamesC = */ 0,
  /***** -colors: colors to use in plots */
  /* colorsP = */ 0,
  /* colors = */ (char**)0,
  /* colorsC = */ 0,
  /***** -bg: background color */
  /* backgroundP = */ 1,
  /* background = */ "red",
  /* backgroundC = */ 1,
  /***** -o: offset to add to all curves to plot */
  /* offsetP = */ 0,
  /* offset = */ (float)0,
  /* offsetC = */ 0,
  /***** -p: probability */
  /* pP = */ 0,
  /* p = */ (float)0,
  /* pC = */ 0,
  /***** -f: frequencies */
  /* fP = */ 1,
  /* f = */ fDefault,
  /* fC = */ 4,
  /***** -negscale: negative scale value */
  /* negscaleP = */ 0,
  /* negscale = */ (float)0,
  /* negscaleC = */ 0,
  /***** -scale: scale value */
  /* scaleP = */ 0,
  /* scale = */ (float)0,
  /* scaleC = */ 0,
  /***** -a: flog quarx flirim poli gam i nabgala */
  /* aP = */ 0,
  /* a = */ (int)0,
  /* aC = */ 0,
  /***** -b: ram dibum gabalabarum deri pum pam */
  /* bP = */ 1,
  /* b = */ bDefault,
  /* bC = */ 3
};



/**********************************************************************/
int
getIntOpt(int argc, char **argv, int i, int *value, int force)
{
  char *end;

  if( ++i>=argc ) goto nothingFound;

  *value = (int)strtol(argv[i], &end, 10);

  /***** check for conversion error */
  if( end==argv[i] ) goto nothingFound;

  /***** check for surplus non-whitespace */
  while( isspace(*end) ) end+=1;
  if( *end ) goto nothingFound;

  return i;

nothingFound:
  if( !force ) return i-1;

  fprintf(stderr, 
	  "%s: missing or malformed integer value after option `%s'\n",
	  Program, argv[i-1]);
    exit(1);
}
/**********************************************************************/
int
getIntOpts(int argc, char **argv, int i, 
	   int **values,
	   int cmin, int cmax)
/*****
  We want to find at least cmin values and at most cmax values.
  cmax==-1 then means infinitely many are allowed.
*****/
{
  int alloced, used;
  char *end;

  if( i+cmin >= argc ) {
    fprintf(stderr, 
	    "%s: option `%s' wants at least %d parameters\n",
	    Program, argv[i], cmin);
    exit(1);
  }

  /***** 
    alloc a bit more than cmin values. It does not hurt to have room
    for a bit more values than cmax.
  *****/
  alloced = cmin + 4;
  *values = (int*)calloc(alloced, sizeof(int));
  if( ! *values ) {
outMem:
    fprintf(stderr, 
	    "%s: out of memory while parsing option `%s'\n",
	    Program, argv[i]);
    exit(1);
  }

  for(used=0; (cmax==-1 || used<cmax) && used+i+1<argc; used++) {
    if( used==alloced ) {
      alloced += 8;
      *values = realloc(*values, alloced*sizeof(int));
      if( !*values ) goto outMem;
    }

    (*values)[used] = (int)strtol(argv[used+i+1], &end, 10);

    /***** check for conversion error */
    if( end==argv[i] ) break;

    /***** check for surplus non-whitespace */
    while( isspace(*end) ) end+=1;
    if( *end ) break;
  }
    
  if( used<cmin ) {
    fprintf(stderr, 
	    "%s: parameter `%s' of `%s' should be an "
	    "integer value\n",
	    Program, argv[i+used+1], argv[i]);
    exit(1);
  }

  return i+used;
}
/**********************************************************************/
int
getFloatOpt(int argc, char **argv, int i, float *value, int force)
{
  char *end;

  if( ++i>=argc ) goto nothingFound;

  *value = (float)strtod(argv[i], &end);

  /***** check for conversion error */
  if( end==argv[i] ) goto nothingFound;

  /***** check for surplus non-whitespace */
  while( isspace(*end) ) end+=1;
  if( *end ) goto nothingFound;

  return i;

nothingFound:
  if( !force ) return i-1;

  fprintf(stderr,
	  "%s: missing or malformed float value after option `%s'\n",
	  Program, argv[i-1]);
  exit(1);
 
}
/**********************************************************************/
int
getFloatOpts(int argc, char **argv, int i, 
	   float **values,
	   int cmin, int cmax)
/*****
  We want to find at least cmin values and at most cmax values.
  cmax==-1 then means infinitely many are allowed.
*****/
{
  int alloced, used;
  char *end;

  if( i+cmin >= argc ) {
    fprintf(stderr, 
	    "%s: option `%s' wants at least %d parameters\n",
	    Program, argv[i], cmin);
    exit(1);
  }

  /***** 
    alloc a bit more than cmin values. It does not hurt to have room
    for a bit more values than cmax.
  *****/
  alloced = cmin + 4;
  *values = (float*)calloc(alloced, sizeof(float));
  if( ! *values ) {
outMem:
    fprintf(stderr, 
	    "%s: out of memory while parsing option `%s'\n",
	    Program, argv[i]);
    exit(1);
  }

  for(used=0; (cmax==-1 || used<cmax) && used+i+1<argc; used++) {
    if( used==alloced ) {
      alloced += 8;
      *values = realloc(*values, alloced*sizeof(float));
      if( !*values ) goto outMem;
    }

    (*values)[used] = (float)strtod(argv[used+i+1], &end);

    /***** check for conversion error */
    if( end==argv[i] ) break;

    /***** check for surplus non-whitespace */
    while( isspace(*end) ) end+=1;
    if( *end ) break;
  }
    
  if( used<cmin ) {
    fprintf(stderr, 
	    "%s: parameter `%s' of `%s' should be a "
	    "floating-point value\n",
	    Program, argv[i+used+1], argv[i]);
    exit(1);
  }

  return i+used;
}
/**********************************************************************/
int
getStringOpt(int argc, char **argv, int i, char **value, int force)
{
  if( ++i>=argc ) {
    fprintf(stderr, "%s: missing string after option `%s'\n",
            Program, argv[i-1]);
    exit(1);
  }
  
  if( !force && argv[i+1][0] == '-' ) return i-1;
  *value = argv[i];
  return i;
}
/**********************************************************************/
int
getStringOpts(int argc, char **argv, int i, 
	   char*  **values,
	   int cmin, int cmax)
/*****
  We want to find at least cmin values and at most cmax values.
  cmax==-1 then means infinitely many are allowed.
*****/
{
  int alloced, used;

  if( i+cmin >= argc ) {
    fprintf(stderr, 
	    "%s: option `%s' wants at least %d parameters\n",
	    Program, argv[i], cmin);
    exit(1);
  }

  alloced = cmin + 4;
    
  *values = (char**)calloc(alloced, sizeof(char*));
  if( ! *values ) {
outMem:
    fprintf(stderr, 
	    "%s: out of memory during parsing of option `%s'\n",
	    Program, argv[i]);
    exit(1);
  }

  for(used=0; (cmax==-1 || used<cmax) && used+i+1<argc; used++) {
    if( used==alloced ) {
      alloced += 8;
      *values = realloc(*values, alloced*sizeof(char*));
      if( !*values ) goto outMem;
    }

    if( argv[used+i+1][0]=='-' ) break;
    (*values)[used] = argv[used+i+1];
  }
    
  if( used<cmin ) {
    fprintf(stderr, 
    "%s: less than %d parameters for option `%s', only %d found\n",
	    Program, cmin, argv[i], used);
    exit(1);
  }

  return i+used;
}
/**********************************************************************/
void
checkIntLower(char *opt, int *values, int count, int max)
{
  int i;

  for(i=0; i<count; i++) {
    if( values[i]<=max ) continue;
    fprintf(stderr, 
	    "%s: parameter %d of option `%s' greater than max=%d\n",
	    Program, i+1, opt, max);
    exit(1);
  }
}
/**********************************************************************/
void
checkIntHigher(char *opt, int *values, int count, int min)
{
  int i;

  for(i=0; i<count; i++) {
    if( values[i]>=min ) continue;
    fprintf(stderr, 
	    "%s: parameter %d of option `%s' smaller than min=%d\n",
	    Program, i+1, opt, min);
    exit(1);
  }
}
/**********************************************************************/
void
checkFloatLower(char *opt, float *values, int count, float max)
{
  int i;

  for(i=0; i<count; i++) {
    if( values[i]<=max ) continue;
    fprintf(stderr, 
	    "%s: parameter %d of option `%s' greater than max=%f\n",
	    Program, i+1, opt, max);
    exit(1);
  }
}
/**********************************************************************/
void
checkFloatHigher(char *opt, float *values, int count, float min)
{
  int i;

  for(i=0; i<count; i++) {
    if( values[i]>=min ) continue;
    fprintf(stderr, 
	    "%s: parameter %d of option `%s' smaller than min=%f\n",
	    Program, i+1, opt, min);
    exit(1);
  }
}
/**********************************************************************/
static void
missingErr(char *opt)
{
  fprintf(stderr, "%s: mandatory option `%s' missing\n",
	  Program, opt);
}
/**********************************************************************/
void
showOptionValues(void)
{
  int i;

  /***** -v: switch on verbose program operation */
  if( !cmd.verboseP ) {
    printf("-v not found.\n");
  } else {
    printf("-v found:\n");
  }

  /***** -title: title of x/y-plot */
  if( !cmd.titleP ) {
    printf("-title not found.\n");
  } else {
    printf("-title found:\n");
    if( !cmd.titleC ) {
      printf("  no values\n");
    } else {
      printf("  value = `%s'\n", cmd.title);
    }
  }

  /***** -xytitles: title of x- and y-axis */
  if( !cmd.xytitlesP ) {
    printf("-xytitles not found.\n");
  } else {
    printf("-xytitles found:\n");
    if( !cmd.xytitlesC ) {
      printf("  no values\n");
    } else {
      printf("  values =");
      for(i=0; i<cmd.xytitlesC; i++) {
        printf(" `%s'", cmd.xytitles[i]);
      }
      printf("\n");
    }
  }

  /***** -plotnames: names of curves to plot */
  if( !cmd.plotnamesP ) {
    printf("-plotnames not found.\n");
  } else {
    printf("-plotnames found:\n");
    if( !cmd.plotnamesC ) {
      printf("  no values\n");
    } else {
      printf("  values =");
      for(i=0; i<cmd.plotnamesC; i++) {
        printf(" `%s'", cmd.plotnames[i]);
      }
      printf("\n");
    }
  }

  /***** -colors: colors to use in plots */
  if( !cmd.colorsP ) {
    printf("-colors not found.\n");
  } else {
    printf("-colors found:\n");
    if( !cmd.colorsC ) {
      printf("  no values\n");
    } else {
      printf("  values =");
      for(i=0; i<cmd.colorsC; i++) {
        printf(" `%s'", cmd.colors[i]);
      }
      printf("\n");
    }
  }

  /***** -bg: background color */
  if( !cmd.backgroundP ) {
    printf("-bg not found.\n");
  } else {
    printf("-bg found:\n");
    if( !cmd.backgroundC ) {
      printf("  no values\n");
    } else {
      printf("  value = `%s'\n", cmd.background);
    }
  }

  /***** -o: offset to add to all curves to plot */
  if( !cmd.offsetP ) {
    printf("-o not found.\n");
  } else {
    printf("-o found:\n");
    if( !cmd.offsetC ) {
      printf("  no values\n");
    } else {
      printf("  value = `%f'\n", cmd.offset);
    }
  }

  /***** -p: probability */
  if( !cmd.pP ) {
    printf("-p not found.\n");
  } else {
    printf("-p found:\n");
    if( !cmd.pC ) {
      printf("  no values\n");
    } else {
      printf("  value = `%f'\n", cmd.p);
    }
  }

  /***** -f: frequencies */
  if( !cmd.fP ) {
    printf("-f not found.\n");
  } else {
    printf("-f found:\n");
    if( !cmd.fC ) {
      printf("  no values\n");
    } else {
      printf("  values =");
      for(i=0; i<cmd.fC; i++) {
        printf(" `%f'", cmd.f[i]);
      }
      printf("\n");
    }
  }

  /***** -negscale: negative scale value */
  if( !cmd.negscaleP ) {
    printf("-negscale not found.\n");
  } else {
    printf("-negscale found:\n");
    if( !cmd.negscaleC ) {
      printf("  no values\n");
    } else {
      printf("  value = `%f'\n", cmd.negscale);
    }
  }

  /***** -scale: scale value */
  if( !cmd.scaleP ) {
    printf("-scale not found.\n");
  } else {
    printf("-scale found:\n");
    if( !cmd.scaleC ) {
      printf("  no values\n");
    } else {
      printf("  value = `%f'\n", cmd.scale);
    }
  }

  /***** -a: flog quarx flirim poli gam i nabgala */
  if( !cmd.aP ) {
    printf("-a not found.\n");
  } else {
    printf("-a found:\n");
    if( !cmd.aC ) {
      printf("  no values\n");
    } else {
      printf("  value = `%d'\n", cmd.a);
    }
  }

  /***** -b: ram dibum gabalabarum deri pum pam */
  if( !cmd.bP ) {
    printf("-b not found.\n");
  } else {
    printf("-b found:\n");
    if( !cmd.bC ) {
      printf("  no values\n");
    } else {
      printf("  values =");
      for(i=0; i<cmd.bC; i++) {
        printf(" `%d'", cmd.b[i]);
      }
      printf("\n");
    }
  }
  if( !cmd.argc ) {
    printf("no remaining parameters in argv\n");
  } else {
    printf("argv =");
    for(i=0; i<cmd.argc; i++) {
      printf(" `%s'", cmd.argv[i]);
    }
    printf("\n");
  }
}
/**********************************************************************/


void
usage(void)
{
  fprintf(stderr, "\
usage: %s [-v] [-title title] [-xytitles xytitles] [-plotnames plotnames] -colors colors [-bg background] [-o offset] [-p p] [-f f] [-negscale negscale] [-scale scale] -a a [-b b] infiles ...\n\
    demonstrate features of the clig-generated command line parser\n\
          -v: switch on verbose program operation\n\
      -title: title of x/y-plot\n\
              1 string value\n\
   -xytitles: title of x- and y-axis\n\
              2 string values\n\
  -plotnames: names of curves to plot\n\
              1 or more string values\n\
     -colors: colors to use in plots\n\
              1...5 string values\n\
         -bg: background color\n\
              1 string value\n\
              default: `red'\n\
          -o: offset to add to all curves to plot\n\
              1 floating-point value\n\
          -p: probability\n\
              1 floating-point value between 0 and 1\n\
          -f: frequencies\n\
              2...10 floating-point values between 0 and 47.11\n\
              default: `2' `4' `8' `16.11'\n\
   -negscale: negative scale value\n\
              1 floating-point value between -oo and 0.0\n\
      -scale: scale value\n\
              1 floating-point value between 0.0 and oo\n\
          -a: flog quarx flirim poli gam i nabgala\n\
              1 integer value\n\
          -b: ram dibum gabalabarum deri pum pam\n\
              3...4 integer values between -10 and 10\n\
              default: `-1' `0' `1'\n\
     infiles: list of input files\n\
              1...10 string values\n
version: 1996-04-07\n\
", Program);
  exit(1);
}
/**********************************************************************/

Cmdline *
parseCmdline(int argc, char **argv)
{
  int i, keep;
  char missingMandatory = 0;

  Program = argv[0];
  for(i=1, cmd.argc=1; i<argc; i++) {
    if( 0==strcmp("-v", argv[i]) ) {
      cmd.verboseP = 1;
      continue;
    }

    if( 0==strcmp("-title", argv[i]) ) {
      cmd.titleP = 1;
      keep = i;
      i = getStringOpt(argc, argv, i, &cmd.title, 1);
      cmd.titleC = i-keep;
      continue;
    }

    if( 0==strcmp("-xytitles", argv[i]) ) {
      cmd.xytitlesP = 1;
      keep = i;
      i = getStringOpts(argc, argv, i, &cmd.xytitles, 2, 2);
      cmd.xytitlesC = i-keep;
      continue;
    }

    if( 0==strcmp("-plotnames", argv[i]) ) {
      cmd.plotnamesP = 1;
      keep = i;
      i = getStringOpts(argc, argv, i, &cmd.plotnames, 1, -1);
      cmd.plotnamesC = i-keep;
      continue;
    }

    if( 0==strcmp("-colors", argv[i]) ) {
      cmd.colorsP = 1;
      keep = i;
      i = getStringOpts(argc, argv, i, &cmd.colors, 1, 5);
      cmd.colorsC = i-keep;
      continue;
    }

    if( 0==strcmp("-bg", argv[i]) ) {
      cmd.backgroundP = 1;
      keep = i;
      i = getStringOpt(argc, argv, i, &cmd.background, 1);
      cmd.backgroundC = i-keep;
      continue;
    }

    if( 0==strcmp("-o", argv[i]) ) {
      cmd.offsetP = 1;
      keep = i;
      i = getFloatOpt(argc, argv, i, &cmd.offset, 1);
      cmd.offsetC = i-keep;
      continue;
    }

    if( 0==strcmp("-p", argv[i]) ) {
      cmd.pP = 1;
      keep = i;
      i = getFloatOpt(argc, argv, i, &cmd.p, 1);
      cmd.pC = i-keep;
      checkFloatLower("-p", &cmd.p, cmd.pC, 1);
      checkFloatHigher("-p", &cmd.p, cmd.pC, 0);
      continue;
    }

    if( 0==strcmp("-f", argv[i]) ) {
      cmd.fP = 1;
      keep = i;
      i = getFloatOpts(argc, argv, i, &cmd.f, 2, 10);
      cmd.fC = i-keep;
      checkFloatLower("-f", cmd.f, cmd.fC, 47.11);
      checkFloatHigher("-f", cmd.f, cmd.fC, 0);
      continue;
    }

    if( 0==strcmp("-negscale", argv[i]) ) {
      cmd.negscaleP = 1;
      keep = i;
      i = getFloatOpt(argc, argv, i, &cmd.negscale, 1);
      cmd.negscaleC = i-keep;
      checkFloatLower("-negscale", &cmd.negscale, cmd.negscaleC, 0.0);
      continue;
    }

    if( 0==strcmp("-scale", argv[i]) ) {
      cmd.scaleP = 1;
      keep = i;
      i = getFloatOpt(argc, argv, i, &cmd.scale, 1);
      cmd.scaleC = i-keep;
      checkFloatHigher("-scale", &cmd.scale, cmd.scaleC, 0.0);
      continue;
    }

    if( 0==strcmp("-a", argv[i]) ) {
      cmd.aP = 1;
      keep = i;
      i = getIntOpt(argc, argv, i, &cmd.a, 1);
      cmd.aC = i-keep;
      continue;
    }

    if( 0==strcmp("-b", argv[i]) ) {
      cmd.bP = 1;
      keep = i;
      i = getIntOpts(argc, argv, i, &cmd.b, 3, 4);
      cmd.bC = i-keep;
      checkIntLower("-b", cmd.b, cmd.bC, 10);
      checkIntHigher("-b", cmd.b, cmd.bC, -10);
      continue;
    }

    if( argv[i][0]=='-' ) {
      fprintf(stderr, "\n%s: unknown option `%s'\n\n",
              Program, argv[i]);
      usage();
    }
    argv[cmd.argc++] = argv[i];
  }/* for i */

  if( !cmd.colorsP ) {
    missingErr("-colors");
    missingMandatory = 1;
  }
  if( !cmd.aP ) {
    missingErr("-a");
    missingMandatory = 1;
  }
  if( missingMandatory ) exit(1);

  cmd.argv = argv+1;
  cmd.argc -= 1;

  if( 1>cmd.argc ) {
    fprintf(stderr, "%s: there should be at least 1 non-option argument(s)\n",
            Program);
    exit(1);
  }
  if( 10<cmd.argc ) {
    fprintf(stderr, "%s: there should be at most 10 non-option argument(s)\n",
            Program);
    exit(1);
  }
  return &cmd;
}

